/* $Id: mdnsclient.c,v 1.1.2.16 2005/09/08 12:27:56 tmh Exp $ */

/***
  This file is part of nss-mdns.
 
  nss-mdns is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.
 
  nss-mdns is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU Lesser General Public
  License along with nss-mdns; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.
***/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef _WIN32
#include "win32/inttypes.h"
#else
#include <stdarg.h> // hpux needs this before inttypes.h
#include <inttypes.h>
#endif
#include <errno.h>
#include <string.h>
#include <stdio.h>
#ifdef _WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif
#endif
#include <fcntl.h>
#include <assert.h>
#ifdef WIN32
#include <time.h>
#else
#include <sys/time.h>
#include <net/if.h>
#include <sys/ioctl.h>
#endif

// OSX doesn't define this
#ifndef SOL_IP
#define SOL_IP IPPROTO_IP
#endif

#ifdef sun
/* Solaris defines this, but not recvmsg, so it's uselss. */
#undef IP_RECVTTL
#endif

#include "mdnsclient.h"
#include "dns.h"
#include "util.h"

#ifdef _WIN32
#define sock_errno WSAGetLastError()
#define EWOULDBLOCK WSAEWOULDBLOCK
#define sock_strerror gai_strerror
#else
#define sock_errno errno
#define closesocket close
#define sock_strerror strerror
#endif

enum
{
	SQ_TXT = 1,
	SQ_SRV = 2
};

#define MDNS_SERV "_services._dns-sd._udp.local"

static const usec_t retry_ms[] = { 200000, 500000, 900000, 1400000, 0 };

static void mdns_mcast_group(struct sockaddr_in *ret_sa) {
    assert(ret_sa);
    
    ret_sa->sin_family = AF_INET;
    ret_sa->sin_port = htons(5353);
    ret_sa->sin_addr.s_addr = inet_addr("224.0.0.251");
}

sock_t mdns_open_socket(void) {
    struct ip_mreq mreq;
    struct sockaddr_in sa;
	sock_t fd = (sock_t)-1;
    u_char ttl;
    int yes;

    mdns_mcast_group(&sa);
        
    if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
        fprintf(stderr, "socket() failed: %s\n", sock_strerror(sock_errno));
        goto fail;
    }
    
    ttl = 255;
    if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, (const char *)&ttl, sizeof(ttl)) < 0) {
        fprintf(stderr, "IP_MULTICAST_TTL failed: %s\n", sock_strerror(sock_errno));
        goto fail;
    }

#ifdef SO_REUSEADDR
    yes = 1;
    if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (const char *)&yes, sizeof(yes)) < 0) {
        fprintf(stderr, "SO_REUSEADDR failed: %s\n", sock_strerror(sock_errno));
        goto fail;
    }
#endif

#ifdef SO_REUSEPORT
    yes = 1;
    if (setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, (const char *)&yes, sizeof(yes)) < 0) {
        fprintf(stderr, "SO_REUSEPORT failed: %s\n", sock_strerror(sock_errno));
        goto fail;
    }
#endif

    memset(&mreq, 0, sizeof(mreq));
    mreq.imr_multiaddr = sa.sin_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);
    sa.sin_addr.s_addr=htons(INADDR_ANY);

    if (bind(fd, (struct sockaddr*) &sa, sizeof(sa)) < 0) {
        fprintf(stderr, "bind() failed: %s\n", sock_strerror(sock_errno));
        goto fail;
    }
    
    if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (const char *)&mreq, sizeof(mreq)) < 0) {
        fprintf(stderr, "IP_ADD_MEMBERSHIP failed: %s\n", sock_strerror(sock_errno));
        goto fail;
    }

#ifdef IP_RECVTTL
	/* Linux specific (I think). */
    if (setsockopt(fd, IPPROTO_IP, IP_RECVTTL, (const char *)&yes, sizeof(yes)) < 0) {
        fprintf(stderr, "IP_RECVTTL failed: %s\n", sock_strerror(sock_errno));
        goto fail;
    }
#endif
   
    if (set_cloexec(fd) < 0) {
        fprintf(stderr, "FD_CLOEXEC failed: %s\n", sock_strerror(sock_errno));
        goto fail;
    }
    
    if (set_nonblock(fd) < 0) {
        fprintf(stderr, "O_ONONBLOCK failed: %s\n", sock_strerror(sock_errno));
        goto fail;
    }

    return fd;

fail:
    if (fd >= 0)
        closesocket(fd);

    return -1;
}

int mdns_close_socket(sock_t sock)
{
	return closesocket(sock);
}

static int send_dns_packet(sock_t fd, struct dns_packet *p)
{
    struct sockaddr_in sa;

    assert(fd >= 0 && p);
    assert(dns_packet_check_valid(p) >= 0);

    mdns_mcast_group(&sa);
       
    for (;;) {
        
        if (sendto(fd, p->data, (int)p->size, 0, (struct sockaddr *)&sa, sizeof(sa)) >= 0)
            break;

        if (sock_errno != EAGAIN) {
	    fprintf(stderr, "sendto() failed: %s\n", sock_strerror(sock_errno));
            return -1;
        }
        
        if (wait_for_write(fd, NULL) < 0)
            return -1;
    }

	return 1;
}

static int recv_dns_packet(sock_t fd, struct dns_packet **ret_packet, struct timeval *end, struct sockaddr_storage *from, int from_len, int *ttl)
{
    struct dns_packet *p= NULL;
    int ret = -1;
    assert(fd >= 0);

    p = dns_packet_new();
   
    for (;;) {
        ssize_t l;
        int r;
        
#ifdef IP_RECVTTL /* In theory anyone that has this also has recvmsg.. */
		struct msghdr msg = {0};
		struct iovec iov[1];
		char control[1024];

		iov[0].iov_base = p->data;
		iov[0].iov_len = sizeof(p->data);
		msg.msg_iov = iov;
		msg.msg_iovlen = 1;
		msg.msg_name = from;
		msg.msg_namelen = from_len;
		msg.msg_control = control;
		msg.msg_controllen = sizeof(control);

		if(( l = recvmsg(fd, &msg, 0)) >= 0)
		{
			struct cmsghdr *cmsg;
			*ttl=255;
			for(cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg,cmsg))
			{
				if(cmsg->cmsg_level == SOL_IP && cmsg->cmsg_type == IP_TTL)
				{
					*ttl = *(int*) CMSG_DATA(cmsg);
					break;
				}
			}

			p->size = (size_t) l;
			*ret_packet = p;
			return 0;
		}
#else
		if ((l = recvfrom(fd, p->data, (int)sizeof(p->data), 0, (struct sockaddr*)from, &from_len)) >= 0)
		{
            p->size = (size_t) l;

			if(ttl) *ttl=255;
            *ret_packet = p;
            return 0;
        }	
#endif
        if (sock_errno != EAGAIN && sock_errno!=EWOULDBLOCK) {
            fprintf(stderr, "recvfrom() failed: %s\n", sock_strerror(sock_errno));
            goto fail;
        }
        
        if ((r = wait_for_read(fd, end)) < 0)
            goto fail;
        else if (r > 0) { /* timeout */
            ret = 1;
            goto fail;
        }
    }

fail:
    if (p)
        dns_packet_free(p);

    return ret;
}

static int send_name_query(sock_t fd, const char *name, int query_ipv4, int query_ipv6) {
    int ret = -1;
    struct dns_packet *p = NULL;
    uint8_t *prev_name = NULL;
    int qdcount = 0;

    assert(fd >= 0 && name && (query_ipv4 || query_ipv6));

    if (!(p = dns_packet_new())) {
        fprintf(stderr, "Failed to allocate DNS packet.\n");
        goto finish;
    }

    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(0, 0, 0, 0, 0, 0, 0, 0, 0, 0));

    if (query_ipv4) {
        if (!(prev_name = dns_packet_append_name(p, name))) {
            fprintf(stderr, "Bad host name\n");
            goto finish;
            
        }
        dns_packet_append_uint16(p, DNS_TYPE_A);
        dns_packet_append_uint16(p, DNS_CLASS_IN);
        qdcount++;
    }
    
    if (query_ipv6) {
        if (!dns_packet_append_name_compressed(p, name, prev_name)) {
            fprintf(stderr, "Bad host name\n");
            goto finish;
        }
        
        dns_packet_append_uint16(p, DNS_TYPE_AAAA);
        dns_packet_append_uint16(p, DNS_CLASS_IN);
        qdcount++;
    }

	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, qdcount);
    
    ret = send_dns_packet(fd, p);
    
finish:
    if (p)
        dns_packet_free(p);

    return ret;
}

static int send_service_query(sock_t fd, const char *service, int flags)
{
    int ret = -1;
    struct dns_packet *p = NULL;
    uint8_t *prev_name = NULL;
    int qdcount = 0;

    assert(fd >= 0);

    if (!(p = dns_packet_new())) {
        fprintf(stderr, "Failed to allocate DNS packet.\n");
        goto finish;
    }

    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(0, 0, 0, 0, 0, 0, 0, 0, 0, 0));

    if (!(prev_name = dns_packet_append_name(p, service))) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
        
    }
    dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
    qdcount++;
    
    if (!(prev_name = dns_packet_append_name(p, service))) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
        
    }

	if(flags&SQ_SRV)
	{
		dns_packet_append_uint16(p, DNS_TYPE_SRV);
		dns_packet_append_uint16(p, DNS_CLASS_IN);
		qdcount++;
	}
	else if(flags&SQ_TXT)
	{
		dns_packet_append_uint16(p, DNS_TYPE_TXT);
		dns_packet_append_uint16(p, DNS_CLASS_IN);
		qdcount++;
	}

	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, qdcount);
    
    ret = send_dns_packet(fd, p);
    
finish:
    if (p)
        dns_packet_free(p);

    return ret;
}

static int domain_cmp(const char *a, const char *b) {
    size_t al, bl;

    al = strlen(a);
    bl = strlen(b);

    if (al > 0 && a[al-1] == '.')
        al --;

    if (bl > 0 && b[bl-1] == '.')
        bl --;

    if (al != bl)
        return al > bl ? 1 : (al < bl ? -1 : 0);

    return strncasecmp(a, b, al);
}

static int process_name_response(sock_t fd, const char *name, usec_t timeout, void (*ipv4_func)(const ipv4_address_t *ipv4, void *userdata), void (*ipv6_func)(const ipv6_address_t *ipv6, void *userdata), void *userdata) {
    struct dns_packet *p = NULL;
    int done = 0;
    struct timeval end;
	struct sockaddr_storage from;
    
    assert(fd >= 0 && name && (ipv4_func || ipv6_func));

    gettimeofday(&end, NULL);
    timeval_add(&end, timeout);

    while (!done) {
        int r,ttl;

        if ((r = recv_dns_packet(fd, &p, &end, &from, (int)sizeof(from), &ttl)) < 0)
            return -1;
        else if (r > 0) /* timeout */
            return 1;

        /* Ignore corrupt packets */
	/* ttl==255 is pure MDNS.  ttl==1 is IETF */
        if ((ttl == 255 || ttl==1) && dns_packet_check_valid_response(p) >= 0) {
            
            for (;;) {
                char pname[256];
                uint16_t type, class;
                uint32_t rr_ttl;
                uint16_t rdlength;
                
                if (dns_packet_consume_name(p, pname, sizeof(pname)) < 0 ||
                    dns_packet_consume_uint16(p, &type) < 0 ||
                    dns_packet_consume_uint16(p, &class) < 0 ||
                    dns_packet_consume_uint32(p, &rr_ttl) < 0 ||
                    dns_packet_consume_uint16(p, &rdlength) < 0) {
                    break;
                }

                /* Remove mDNS cache flush bit */
                class &= ~DNS_CLASS_FLUSH;
                
                if (ipv4_func &&
                    type == DNS_TYPE_A &&
                    class == DNS_CLASS_IN &&
                    !domain_cmp(name, pname) &&
                    rdlength == sizeof(ipv4_address_t)) {

                    ipv4_address_t ipv4;
                    
                    if (dns_packet_consume_bytes(p, &ipv4, sizeof(ipv4)) < 0)
                        break;

                    ipv4_func(&ipv4, userdata);
                    done = 1;
                    
                }
                else
                    if (ipv6_func &&
                            type == DNS_TYPE_AAAA &&
                            class == DNS_CLASS_IN &&
                            !domain_cmp(name, pname) &&
                            rdlength == sizeof(ipv6_address_t)) {

                    ipv6_address_t ipv6;
                    
                    if (dns_packet_consume_bytes(p, &ipv6, sizeof(ipv6_address_t)) < 0)
                        break;
                    
                    ipv6_func(&ipv6, userdata);
                    done = 1;
                }
                else 
				{

                    /* Step over */
                    
                    if (dns_packet_consume_seek(p, rdlength) < 0)
                        break;
                }
            }
        }

        if (p)
            dns_packet_free(p);
    }

    return 0;
}

static int process_service_response(sock_t fd, const char *prefix, usec_t timeout, struct service_handlers *service_func, void *userdata)
{
    struct dns_packet *p = NULL;
    int done = 0;
    struct timeval end;
	struct sockaddr_storage from;
    
    assert(fd >= 0 && service_func);

	gettimeofday(&end, NULL);
	timeval_add(&end, timeout);
    while (!done) {
        int r,ttl;

        if ((r = recv_dns_packet(fd, &p, &end, &from, (int)sizeof(from), &ttl)) < 0)
            return -1;
        else if (r > 0) /* timeout */
            return 1;

        /* Ignore corrupt packets */
        if ((ttl == 255 || ttl == 1) && dns_packet_check_valid_response(p) >= 0) {

			/* Reset timeout */
			gettimeofday(&end, NULL);
			timeval_add(&end, timeout);

            for (;;) {
                char pname[256];
                uint16_t type, class;
                uint32_t rr_ttl;
                uint16_t rdlength;
                
                if (dns_packet_consume_name(p, pname, sizeof(pname)) < 0 ||
                    dns_packet_consume_uint16(p, &type) < 0 ||
                    dns_packet_consume_uint16(p, &class) < 0 ||
                    dns_packet_consume_uint32(p, &rr_ttl) < 0 ||
                    dns_packet_consume_uint16(p, &rdlength) < 0) {
                    break;
                }

                /* Remove mDNS cache flush bit */
                class &= ~DNS_CLASS_FLUSH;
                
				if(type == DNS_TYPE_PTR && class == DNS_CLASS_IN && !domain_cmp(MDNS_SERV, pname))
				{
					char name[256];

					if (dns_packet_consume_name(p, name, sizeof(name)) < 0)
						break;

					if(!prefix || !strncmp(name,prefix,sizeof(prefix)))
						send_service_query(fd, name, SQ_SRV);
				}
				else if(type == DNS_TYPE_PTR && class == DNS_CLASS_IN)
				{
					char name[256];

					if (dns_packet_consume_name(p, name, sizeof(name)) < 0)
						break;
                
					if(service_func->service_name_func)
						service_func->service_name_func(name, userdata);
				}
				else if(type == DNS_TYPE_SRV && class == DNS_CLASS_IN)
				{
					uint16_t priority,weight,port;
					char target[256];
					// srv record
					// Priority Weight Port Target

					dns_packet_consume_uint16(p,&priority);
					dns_packet_consume_uint16(p,&weight);
					dns_packet_consume_uint16(p,&port);
					dns_packet_consume_name(p,target,sizeof(target));
					
					if(strcmp(target,".") && service_func->service_srv_func)
						service_func->service_srv_func(pname, priority, weight, port, target, userdata);

					if(service_func->service_ipv4_func || service_func->service_ipv6_func)
						send_name_query(fd, target, !!service_func->service_ipv4_func, !!service_func->service_ipv6_func);
					if(service_func->service_txt_func)
						send_service_query(fd, pname, SQ_TXT);
				}
				else if(type == DNS_TYPE_TXT && class == DNS_CLASS_IN)
				{
					char buf[1024];

					if (dns_packet_consume_text(p, buf, sizeof(buf)) < 0)
						break;

					if(service_func->service_txt_func)
						service_func->service_txt_func(pname, buf, userdata);
				}
                else if (type == DNS_TYPE_A &&
						class == DNS_CLASS_IN &&
						rdlength == sizeof(ipv4_address_t))
				{
                    ipv4_address_t ipv4;
                    
                    if (dns_packet_consume_bytes(p, &ipv4, sizeof(ipv4)) < 0)
                        break;

					if(service_func->service_ipv4_func)
						service_func->service_ipv4_func(pname, &ipv4, userdata);                   
                }
                else if (type == DNS_TYPE_AAAA &&
                         class == DNS_CLASS_IN &&
                         rdlength == sizeof(ipv6_address_t))
				{
                    ipv6_address_t ipv6;
                    
                    if (dns_packet_consume_bytes(p, &ipv6, sizeof(ipv6)) < 0)
                        break;
                    
					if(service_func->service_ipv6_func)
						service_func->service_ipv6_func(pname, &ipv6, userdata);                   
                }
					else
				{
                    /* Step over */
                    
                    if (dns_packet_consume_seek(p, rdlength) < 0)
                        break;
                }
            }
        }

        if (p)
            dns_packet_free(p);
    }

    return 0;
}

int mdns_query_name(sock_t fd, const char *name, void (*ipv4_func)(const ipv4_address_t *ipv4, void *userdata), void (*ipv6_func)(const ipv6_address_t *ipv6, void *userdata), void *userdata) 
{
	const usec_t *timeout = retry_ms;
    assert(fd >= 0 && name && (ipv4_func || ipv6_func));

    while (*timeout > 0) {
        int n;
        
        if (send_name_query(fd, name, !!ipv4_func, !!ipv6_func) < 0)
            return -1;

        if ((n = process_name_response(fd, name, *timeout, ipv4_func, ipv6_func, userdata)) < 0)
            return -1;

        if (n == 0)
            return 0;

        /* Timeout */

        timeout++;
    }

    return -1;
}

int mdns_query_services(sock_t fd, const char *prefix, struct service_handlers *service_func, void *userdata, usec_t timeout)
{
	int n;
    assert(fd >= 0 && service_func);

    if(!timeout)
		timeout = 2000000;

	if (send_service_query(fd, MDNS_SERV, 0) < 0)
        return -1;

    if ((n = process_service_response(fd, prefix, timeout, service_func, userdata)) < 0)
        return -1;

    if (n == 0)
       return 0;

	/* Timeout.. normal */
    return 0;
}

static int send_reverse_query(sock_t fd, const char *name) {
    int ret = -1;
    struct dns_packet *p = NULL;

    assert(fd >= 0 && name);

    if (!(p = dns_packet_new())) {
        fprintf(stderr, "Failed to allocate DNS packet.\n");
        goto finish;
    }

    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(0, 0, 0, 0, 0, 0, 0, 0, 0, 0));

    if (!dns_packet_append_name(p, name)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }

    dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);

    dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 1);
    
    ret = send_dns_packet(fd, p);
    
finish:
    if (p)
        dns_packet_free(p);

    return ret;
}

static int process_reverse_response(sock_t fd, const char *name, usec_t timeout, void (*name_func)(const char *name, void *userdata), void *userdata) {
    struct dns_packet *p = NULL;
    int done = 0;
    struct timeval end;
	struct sockaddr_storage from;

    assert(fd >= 0 && name && name_func);
    
    gettimeofday(&end, NULL);
    timeval_add(&end, timeout);

    while (!done) {
        int r, ttl;

        if ((r = recv_dns_packet(fd, &p, &end, &from, (int)sizeof(from), &ttl)) < 0)
            return -1;
        else if (r > 0) /* timeout */
            return 1;

        /* Ignore corrupt packets */
        if ((ttl == 255 || ttl == 1 ) && dns_packet_check_valid_response(p) >= 0) {
            
            for (;;) {
                char pname[256];
                uint16_t type, class;
                uint32_t rr_ttl;
                uint16_t rdlength;
                
                if (dns_packet_consume_name(p, pname, sizeof(pname)) < 0 ||
                    dns_packet_consume_uint16(p, &type) < 0 ||
                    dns_packet_consume_uint16(p, &class) < 0 ||
                    dns_packet_consume_uint32(p, &rr_ttl) < 0 ||
                    dns_packet_consume_uint16(p, &rdlength) < 0) {
                    break;
                }

                /* Remove mDNS cache flush bit */
                class &= ~DNS_CLASS_FLUSH;
                
                if (type == DNS_TYPE_PTR &&
                    class == DNS_CLASS_IN &&
                    !domain_cmp(name, pname)) {

                    char rname[256];

                    if (dns_packet_consume_name(p, rname, sizeof(rname)) < 0)
                        break;

                    name_func(rname, userdata);
                    done = 1;
                    
                } else {

                    /* Step over */
                    
                    if (dns_packet_consume_seek(p, rdlength) < 0)
                        break;
                }
            }
        }

        if (p)
            dns_packet_free(p);
    }

    return 0;
}

static int query_reverse(sock_t fd, const char *name, void (*name_func)(const char *name, void *userdata), void *userdata) {
    const usec_t *timeout = retry_ms;
    assert(fd >= 0 && name && name_func);

    while (*timeout > 0) {
        int n;
        
        if (send_reverse_query(fd, name) <= 0) /* error or no interface to send data on */
            return -1;

        if ((n = process_reverse_response(fd, name, *timeout, name_func, userdata)) < 0)
            return -1;

        if (n == 0)
            return 0;

        /* Timeout */

        timeout++;
    }

    return -1;
}

int mdns_query_ipv4(sock_t fd, const ipv4_address_t *ipv4, void (*name_func)(const char *name, void *userdata), void *userdata) {
    char name[256];
    assert(fd >= 0 && ipv4 && name_func);

    snprintf(name, sizeof(name), "%u.%u.%u.%u.in-addr.arpa", ipv4->address[0],ipv4->address[1],ipv4->address[2],ipv4->address[3]);

    return query_reverse(fd, name, name_func, userdata);
}

int mdns_query_ipv6(sock_t fd, const ipv6_address_t *ipv6, void (*name_func)(const char *name, void *userdata), void *userdata) {
    char name[256];
    assert(fd >= 0 && ipv6 && name_func);

    snprintf(name, sizeof(name), "%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.ip6.arpa",
             ipv6->address[15] & 0xF, ipv6->address[15] >> 4,
             ipv6->address[14] & 0xF, ipv6->address[14] >> 4,
             ipv6->address[13] & 0xF, ipv6->address[13] >> 4,
             ipv6->address[12] & 0xF, ipv6->address[12] >> 4,
             ipv6->address[11] & 0xF, ipv6->address[11] >> 4,
             ipv6->address[10] & 0xF, ipv6->address[10] >> 4,
             ipv6->address[9] & 0xF, ipv6->address[9] >> 4,
             ipv6->address[8] & 0xF, ipv6->address[8] >> 4,
             ipv6->address[7] & 0xF, ipv6->address[7] >> 4,
             ipv6->address[6] & 0xF, ipv6->address[6] >> 4,
             ipv6->address[5] & 0xF, ipv6->address[5] >> 4,
             ipv6->address[4] & 0xF, ipv6->address[4] >> 4,
             ipv6->address[3] & 0xF, ipv6->address[3] >> 4,
             ipv6->address[2] & 0xF, ipv6->address[2] >> 4,
             ipv6->address[1] & 0xF, ipv6->address[1] >> 4,
             ipv6->address[0] & 0xF, ipv6->address[0] >> 4);
    
    return query_reverse(fd, name, name_func, userdata);
}

